(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2008,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: genr.ml,v 1.46 2008/02/11 23:05:59 weis Exp $ *)

(** {3 Generation of construction functions for a type definition.} *)

open Parsetree
open Code
open Genr_base
open Relation
;;

(** The global generation procedure: it generates the whole set of construction
  function definitions from a given type definition. *)
let get_varyadic_preamble, set_varyadic_preamble =
  let varyadic_preamble = ref false in
  (fun () -> !varyadic_preamble),
  (fun () -> varyadic_preamble := true)
;;

let genr_function_constant fdef =
  let g = fdef.fd_generator in
  [Genr_base.construction_function g, [], Constant g.pgen_desc]
;;

let genr_function_n_ary fdef =
  let g = fdef.fd_generator in
  let args = [Tuple (Code.genr_args "x" fdef.fd_arity)] in
  let body = Genr_base.generator g args in
  [construction_function g, args, body]
;;

let genr_function fdef =
  match fdef.fd_arity with
  | 0 -> genr_function_constant fdef
  | 1 ->
    if Check.is_varyadic_gen fdef.fd_generator
      then set_varyadic_preamble ();
    Genr_unary.genr_function fdef
  | 2 -> Genr_binary.genr_function fdef
  | _ -> Genr_nary.genr_function fdef
;;

let genr_record tname fields =
  let genr_field lab_name =
    (Longident.Lident lab_name, make_Constant lab_name) in
  [Longident.Lident tname, List.map make_Constant fields,
   Record (List.map genr_field fields)]
;;

let genr_abbrev tname _trep relations =
  (* We must generate a Module.make_tname function and a Module.from_tname
     function.

     The Module.make_tname function injects value of type [trep] into type
     [tname]; therefore, it must use the relations defined with type [tname].
     Module.make_tname is generated here.

     The Module.from_tname function just projects values from type [tname] to
     type [trep]; therefore, it is trivial and just a casting identity
     primitive.

     Module.from_tname must appear both in the interface and in the
     implementation of the module. *)
  match Relation.rewrite relations with
  | [] -> [ Longident.Lident ("make_" ^ tname), [x], x]
  | rls ->
    let clauses =
      List.map
        (fun (pat, e) ->
         match e.pexp_desc with
         | Pexp_when (whe, e) ->
           Clause_when
             (Ocaml_pattern pat, Ocaml_expression whe,
              Ocaml_expression e)
         | _ ->
           Clause (Ocaml_pattern pat, Ocaml_expression e)) rls in
    [Longident.Lident ("make_" ^ tname), [], Function clauses]
;;

let has_relations rels =
  match rels.prels_desc with
  | Prels_none
  | Prels_commented _
  | Prels_begend [] -> false
  | Prels_begend _ -> true
;;

let has_relations_cdef (_s, _cts, rels, _loc) = has_relations rels
;;

let has_relations_cdefs = List.exists has_relations_cdef
;;

let genr_functions_of_type_declaration (tname, td) =

  match td.ptype_kind, td.ptype_private with

  (* Public types: we do not generate construction functions. *)

  (* Regular sum type definition. *)
  | Ptype_variant cdefs, Asttypes.Public ->
    if has_relations_cdefs cdefs then
      Check.error td.ptype_loc (Check.Public_type_with_relations tname)
    else []

  (* Regular record type definition. *)
  | Ptype_record _fdefs, Asttypes.Public -> []

  (* Regular abbreviation type definition. *)
  | Ptype_abstract relations, Asttypes.Public ->
    if has_relations relations then
      Check.error td.ptype_loc (Check.Public_type_with_relations tname)
    else []

  (* Private types: we must generate construction functions. *)

  (* Private sum type definition: the true thing ... *)
  | Ptype_variant cdefs, Asttypes.Private ->
    let fdefs = Complete.genr_function_definitions cdefs in
    let gen_fun = List.flatten (List.map genr_function fdefs) in
    if get_varyadic_preamble ()
    then
      Genr_vary.common_aux_functions @ gen_fun
    else
      gen_fun

  (* Private record type definition. *)
  (* We generate a trivial construction function. *)
  | Ptype_record fdefs, Asttypes.Private ->
    let fields =
      List.map (fun (lab_name, _mut_flag, _ct, _rels, _loc) -> lab_name)
        fdefs in
    genr_record tname fields

  (* Private abbreviation type definition. *)
  (* When sig_item would be available in relations,
     we could (trivially) generate clever injection functions.
     For the time being, we generate a trivial [make] identity construction
     function. *)
  (* We generate a projection function. *)
  | Ptype_abstract relations, Asttypes.Private ->
    match td.ptype_manifest with
    | None -> []
    | Some trep -> genr_abbrev tname trep relations
;;

let genr_projection (tname, tdecl) =
  match tdecl.ptype_kind, tdecl.ptype_private with
  | Ptype_abstract _relations, Asttypes.Private ->
      begin match tdecl.ptype_manifest with
      | None -> []
      | Some trep -> [Genr_base.genr_projection (tname, tdecl) trep]
      end
  | Ptype_abstract _, Asttypes.Public
  | Ptype_variant _, _
  | Ptype_record _, _ -> []
;;

let genr_functions ntds =
  let gen_fun_list =
    List.flatten
     (List.map genr_functions_of_type_declaration ntds) in

  List.flatten (List.map genr_projection ntds) @

  topletrec gen_fun_list :: List.map Genr_base.genr_eq ntds
;;
