(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2008,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: genr_vary.ml,v 1.95 2008/02/12 22:13:02 blanqui Exp $ *)

(** {3 Construction functions for generators with variable arity.} *)

(** We treat the case of unary generators whose unique argument is a list
    that in fact is the variable length list of arguments of an associative
    binary operation.

    We hence generate a unary function for associative generators with a list
    argument. *)

open Relation
open Code
open Parsetree
open Genr_base
open Longident
;;

(* Auxilliaries to generate Code values. *)

(* Usual list operations generating Code values. *)
let cons x y = infix_apply x (Lident "::") y
and nil = make_Constant "[]"
;;

(* Usual arithmetic functions for Code values. *)
let times x y = infix_name x "*" y
and div x y = infix_name x "/" y

(* Usual boolean functions for Code values. *)
let bool_and x y = infix_name x "&&" y

(* Usual comparison function for Code values. *)
let compare x y = apply2 (Ldot (Lident "Pervasives", "compare")) x y
and lt x y = infix_name x "<" y
and eq x y = infix_name x "=" y
and leq x y = infix_name x "<=" y
;;

(* We do not use the system compare (Pervasives.compare) to let the user
   define its own comparison for the relational type at hand. *)
let ident_compare = Lident "compare"
and ident_neq = Lident "(<>)"
;;

(* Usual constants for Code values. *)
let int_zero = make_Constant "0"
and int_two = make_Constant "2"
and bool_true = make_Constant "true"
and bool_false = make_Constant "false"
;;

(* Create a Code ident for the Caml List module. *)
let list_ident s = Ldot (Lident "List", s)
;;

(* Useful idents. *)
let ident_l = Lident "l"
and ident_ltemp = Lident "ltemp"
and ident_accu = Lident "accu"
and ident_aux = Lident "aux"
and ident_mem = Lident "mem"
and ident_x = Lident "x"
and ident_y = Lident "y"
and ident_z = Lident "z"
and ident_t = Lident "t"
and ident_m = Lident "m"
and ident_n = Lident "n"
and ident_dist = Lident "dist"
;;

(* Idents of basic functions *)
let ident_pred = Lident "pred"
;;

(* Idents generated by this module. *)
let ident_delete_first = Lident "delete_first"
and ident_unique_args = Lident "unique_args"
and ident_take = Lident "take"
and ident_drop = Lident "drop"
and ident_nilpotence = Lident "nilpotence"
and ident_distrib_args = Lident "distrib_args"
and ident_inverse_neutral = Lident "inverse_neutral"
and ident_distributive_inverse = Lident "distributive_inverse"
;;

let absorbent_in = prefix_construction_function "absorbent"
and neutral_in sd = 
  let prefix = match sd with
    | Left -> "neutral_left"
    | Right -> "neutral_right"
    | Both -> "neutral" in
    prefix_construction_function prefix
and idempotent_in = prefix_construction_function "idempotent"
and distrib_args_in = prefix_construction_function "distrib_args"
and distrib_in = prefix_construction_function2 "distributive"
and nilpotent_in = prefix_construction_function "nilpotent"
and inverse_in = prefix_construction_function "remove_inv"
and flatten_in = prefix_construction_function "flatten"
and insert_in =  prefix_construction_function "insert"
and return_in = prefix_construction_function "return"
and distrib_inv_in = prefix_construction_function "distrib_inv"
and rewrite_in = prefix_construction_function "rewrite"
;;

(* Useful Code ident expressions for this module. *)
let l = Constant ident_l
and aux = Constant ident_aux
and accu = Constant ident_accu
and m = Constant ident_m
and n = Constant ident_n
(* Expressions x, y, z, t, u, are already defined in Code. *)
;;

(* Useful function from the Caml List module. *)
let list_mem = apply2 (list_ident "mem")
and list_length = apply1 (list_ident "length")
and list_append = apply2 (list_ident "append")
and list_map = apply2 (list_ident "map")
and list_rev = apply1 (list_ident "rev")
and list_sort = apply2 (list_ident "sort")
and list_fold_right = apply3 (list_ident "fold_right")
and list_fold_left = apply3 (list_ident "fold_left")
and list_filter = apply2 (list_ident "filter")
;;

(* Functions defined in this module *)
let take = apply2 ident_take
and drop = apply2 ident_drop
;;

let id_not_found = "Not_found"
;;
let exc_not_found = make_Constant id_not_found
and raise_not_found = raise_error id_not_found
;;

(* First, some auxiliary functions (name, [args], code) *)

let genr_mem =
  ident_mem,
  [x],
  Letrec (ident_aux, [l],
      Match (l, [
        Clause (nil, bool_false);
        Clause (cons y l,
          Match (compare x y, [
            Clause
              (int_zero, bool_true);
            Clause_when
              (t, lt t int_zero, bool_false);
            Clause
              (underscore, apply1 ident_aux l);
          ]));
      ]),
    aux)
;;

let genr_unique_args =
  ident_unique_args,
  [l],
  Letrec (ident_aux, [accu; l],
         Match (l, [
           Clause
             (nil, accu);
           Clause
             (cons x l,
              If (list_mem x accu,
                (apply2 ident_aux accu l),
                (apply2 ident_aux (cons x accu) l)));
         ]),
       apply2 ident_aux nil l)
;;

let genr_delete_first =

  ident_delete_first, [z; l],
    Match (l, [
      Clause_when
        (cons x l, eq x z, l);
      Clause
        (cons x l,
         cons x (apply2 ident_delete_first z l));
      Clause (underscore, raise_not_found);
    ])
;;

let genr_take =
  ident_take, [n; l],
  Match (Tuple [n;l],
	 [Clause (Tuple [int_zero; underscore], nil);
	  Clause (Tuple [underscore; nil], nil);
	  Clause (Tuple [n; cons x m], 
		  cons x (apply2 ident_take (apply1 ident_pred n) m))])
;;

let genr_drop =
  ident_drop, [n; l],
  Match (Tuple [n;l],
	 [Clause (Tuple [int_zero; l], l);
	  Clause (Tuple [underscore; nil], nil);
	  Clause (Tuple [n; cons underscore m], 
		  apply2 ident_drop (apply1 ident_pred n) m)])
;;


(* A bunch of normalization function for some keywords *)

let rec genr_rewrite fdef rewrite_rels =
  if List.length rewrite_rels > 0 then
    let fname = rewrite_in fdef.fd_generator in
    let precode = (fname, [l], Match (l, genr_rewrite_clauses l rewrite_rels)) in
    let code = apply1 fname in
      ([precode], [code])
  else ([], [])

and genr_rewrite_clauses l rewrite_rels =
  let default_clause = Clause (underscore, l) in
  (List.map genr_rewrite_clause rewrite_rels) @ [default_clause]

and genr_rewrite_clause (p, e) =
  match Genr_expression.genr_rule p e with
    | Clause (p, e) -> Clause (p, List [e])
    | Clause_when (p, w, e) -> Clause_when (p, w, List [e])
;;

let rec genr_return g rewrite_rels =
  let fname = return_in g in
  let precode = (fname, [l], Match (l, genr_return_clauses g l rewrite_rels)) in
  let code = apply1 fname in
  ([precode], [code])

and genr_return_clauses g l rewrite_rels =
  let strip_gen_clause = Clause (cons x nil, x) in
  let default_clause = Clause (underscore, generator1 g l) in
    strip_gen_clause :: 
      List.map (genr_recursive_clause g l) rewrite_rels  @
      [default_clause]

and genr_recursive_clause g l (p, e) =
  match Genr_expression.genr_rule p e with
    | Clause (p, _) -> Clause (underscored_pattern p, construction1 g l)
    | Clause_when (p, w, _) -> Clause_when (p, w, construction1 g l)
;;

(* Auxiliary functions for normalizing *)

(* Normalizing with side.
   Right -> normalize everything but the first element
   Left -> normalize everything but the last element *)

let apply_ident_on_right fun_ident l =
  Match (l,
	 [Clause (cons x (cons y m),
		  cons x (apply1 fun_ident (cons y m)));
	  Clause (underscore, l)])

let apply_on_right fun_ident fun_def arg =
  let id_all = prefix_ident "all" fun_ident in
  Letrec (id_all, [l], fun_def id_all l,
	  apply_ident_on_right id_all arg)

let apply_on_left fun_ident fun_def arg =
  let id_all = prefix_ident "all" fun_ident in
  let id_right = prefix_ident "right" fun_ident in
    Letrec (id_all, [l], fun_def id_all l,
	    Let (id_right, [m], apply_ident_on_right id_all m,
		 list_rev (apply1 id_right (list_rev arg))))

let apply_on_side sd fun_ident fun_def arg =
  match sd with
    | Left -> apply_on_left fun_ident fun_def arg
    | Right -> apply_on_right fun_ident fun_def arg
    | Both -> fun_def fun_ident arg
;;

(* Adding an element to a list.
   If the constructor is commutative, insert it keeping the list sorted. *)
let add_element fdef list elem =
  let add =
    if Relation.has_commutative fdef.fd_relations
    then apply2 (insert_in fdef.fd_generator)
    else cons in
  add elem list
  

(* When replacing some elements by another one (e.g. nilpotence, inverse),
   we must add it to the list where they have been eliminated.
   If the replacement is the neutral element, do not add it.
   If the constructor is commutative, insert it keeping the list sorted. *)
let add_replacement fdef list replacing_gen =
  let rls = fdef.fd_relations in
  if is_neutral_gen replacing_gen rls
  then list
  else add_element fdef list (construction0 replacing_gen) 
;;

(* Replacing a pair of elements by another one  (e.g inverse, nilpotence) 
 (patt_i, exp_i) are each of the elements to find, in both pattern and exp form *)
let rec find_and_replace fdef rec_name (patt1, exp1) (patt2, exp2) repl arg =
  let repl_clause = replace_clause fdef rec_name repl in
  let clauses =
    if patt1 = patt2
    then [repl_clause patt1 exp1 exp2;
	  default_clause arg]
    else [repl_clause patt1 exp1 exp2;
	  repl_clause patt2 exp2 exp1;
	  default_clause arg] in
  Match (arg, clauses)

and replace_clause fdef rec_name repl patt1 exp1 exp2 =
  let ident_l_rec = Lident "l_rec" in
  let l_rec = Constant ident_l_rec in
  let insert_ident = insert_in fdef.fd_generator in
  Clause (cons patt1 l,
	  Let (
	    ident_l_rec, [], apply1 rec_name l,
	    Try (
	      add_replacement fdef (apply2 ident_delete_first exp2 l_rec) repl,
	      [Clause (exc_not_found,
		       apply2 insert_ident exp1 l_rec)]
	    )))

and default_clause arg = Clause (underscore, arg)
;;


(* The result of removing elements from a list could be the empty list.
   In that case, replace it by a singleton with the indicated element
   (e.g. the neutral element). *)

let element_if_empty elem list =
  Match (list,
	 [Clause (nil, cons (construction0 elem) nil);
	  Clause (underscore, list)])
;;



(* Normalizing functions to apply according to relations *)

(* Sorting (commutativity) *)
let genr_insert fg lidopt =
  let lt, _eq, _gt = genr_comparisons lidopt in
  let insert_ident = insert_in fg in
  (insert_ident, [x; l],
   Match (l, [
     Clause (nil, cons x nil);
     Clause_when (cons y m, lt x y, cons x l);
     Clause (cons y m, cons y (apply2 insert_ident x m)
     )
   ]))
;;

let genr_sort fg lidopt =
  let comp_id = comparison_function lidopt  in
  ([genr_insert fg lidopt], [list_sort comp_id])
;;

(* Flattening (associativity) *)
let genr_flatten_step fg l x accu  =
   Match (x, [
      Clause
        (pattern fg [l],
         list_append l accu);
      Clause
        (underscore,
         cons x accu);
    ])

let genr_flatten_body fg l =
  let id_flatten_step = Lident "flatten_step" in
  Let (id_flatten_step, [x; accu], genr_flatten_step fg m x accu,
      list_fold_right (Constant id_flatten_step) l nil)

let genr_flatten fg =
  let id_flatten = flatten_in fg in
 ([(id_flatten, [l], genr_flatten_body fg l)],
  [apply1 id_flatten])
;;

(* Neutral *)

let filter_not_neutral neutr args =
  list_filter (apply1 ident_neq (generator0 neutr)) args
;;

let remove_all_neutral neutr _rec_name arg = filter_not_neutral neutr arg
;;

let remove_neutral_side sd neutr arg =
  let ident_remove_neutral = Lident "remove_neutral" in
  apply_on_side sd ident_remove_neutral (remove_all_neutral neutr) arg
;;

let genr_rem_neutral fdef sd neutr =
  let fname = neutral_in sd fdef.fd_generator in
  let remove_neutral x =
    if Relation.has_commutative fdef.fd_relations
    then remove_all_neutral neutr fname x
    else remove_neutral_side sd neutr x in
  let ident_no_neutral = Lident "no_neutral_list" in
  let no_neutral = Constant ident_no_neutral in
  let genr_neutral_norm =
    fname, [x],
    Let (ident_no_neutral, [], remove_neutral x,
	 element_if_empty neutr no_neutral) in
    ([genr_neutral_norm], [apply1 fname] )
;;

(* Idempotence and Nilpotence *)

(* Idempotence and nilpotence for associative but
   not commutative constructors: find duplicate subchains of any length.
   For example, find subchain [b, c] in [a, b, c, b, c, d] *)
let do_recursively_with_consecutives rec_name l x m n do_if_consecutive =
  let ident_next = Lident "next" in
  let next = Constant ident_next in
  Let (ident_next, [], drop n l,
       If (eq (take n l) (take n next),
	   apply2 rec_name (do_if_consecutive next n) n,
	   cons x (apply2 rec_name m n)))

let find_consecutive_n fname l n do_if_consecutive =
  Match (l,
	 [Clause (nil, nil);
	  Clause_when (cons x m, leq (times int_two n) (list_length l),
		       do_recursively_with_consecutives fname l x m n do_if_consecutive);
	  Clause (l, l)])

let find_all_consecutive fname sd l do_when_found =
  let consecutive_n_ident = Lident "remove_consecutive_n" in
  let consecutive_ident = Lident "remove_consecutive_up_to" in
  let remove_consecutive_clause =
    Function [Clause (int_zero, l);
	      Clause (n, apply2 consecutive_ident
			(apply2 consecutive_n_ident l n) (apply1 ident_pred n))] in
  let remove_many_consecutive_def _ l =
    Letrec (consecutive_n_ident, [l;n],
	    find_consecutive_n consecutive_n_ident l n do_when_found,
	    Letrec (consecutive_ident, [l], remove_consecutive_clause,
		    apply2 consecutive_ident l (div (list_length l) int_two))) in
  apply_on_side sd fname remove_many_consecutive_def l
;;
	 
(* Idempotence and nilpotence for neither associative
   nor commutative constructors:
   remove single duplicates that are next to each other *)
let find_single_consecutive fname sd l do_if_consecutive =
  let fun_def rec_name l =
    Match (l, 
	   [Clause (nil, nil);
	    Clause (cons x nil, cons x nil);
	    Clause_when (cons x (cons y m),
			 eq x y,
			 (apply1 rec_name (do_if_consecutive y m)));
	    Clause (cons x (cons y m),
		    cons x (apply1 rec_name (cons y m)))]) in
  apply_on_side sd fname fun_def l
;;

(* Idempotence: replace two consecutives by one *)
let drop_one_consecutive dropped_list _ = dropped_list

let idempotent_consecutive_many sd l =
  let ident = Lident "remove_consecutive_duplicates" in
  find_all_consecutive ident sd l drop_one_consecutive

let drop_one_single_consecutive y m = cons y m

let idempotent_consecutive sd l =
  let ident = Lident "remove_consecutive_duplicates" in
  find_single_consecutive ident sd l drop_one_single_consecutive

let genr_idempotent fdef sd =
  let fname = idempotent_in fdef.fd_generator in
  let idempotent_function =
    if Relation.has_commutative fdef.fd_relations
    then apply1 ident_unique_args
    else
      if Relation.has_associative fdef.fd_relations
      then idempotent_consecutive_many sd
      else idempotent_consecutive sd in
  ([fname, [l], idempotent_function l],
   [apply1 fname])


(* Nilpotence: replace two consecutives by indicated element. *)

let replace_consecutives g dropped_list n = cons (construction0 g) (drop n dropped_list)

let nilpotent_consecutive_many sd g l =
  let ident = Lident "replace_consecutive_duplicates" in
  find_all_consecutive ident sd l (replace_consecutives g)

let replace_single_consecutives g _y m = cons (construction0 g)  m

let nilpotent_consecutive fname sd g l =
  find_single_consecutive fname sd l (replace_single_consecutives g)

let genr_nilpotent fdef sd g =
  let fname = nilpotent_in fdef.fd_generator in
  let nilpotent_ident = prefix_ident "do" fname in
  let nilpotent_function =
    if Relation.has_commutative fdef.fd_relations
    then find_and_replace fdef nilpotent_ident (x, x) (x, x) g
    else
      if Relation.has_associative fdef.fd_relations
      then nilpotent_consecutive_many sd g
      else nilpotent_consecutive nilpotent_ident sd g in
  let main_function l =
    Letrec (nilpotent_ident, [l], nilpotent_function l,
	    element_if_empty g (apply1 nilpotent_ident l)) in
  ([fname, [l], main_function l],
   [apply1 fname])
;;

(* Inverse *)
let rec aux_remove_consec_inverse fdef rec_name accum arg invg repl sd =
  let rec_step = apply2 rec_name in
  let add_elem x l1 l2 = rec_step (add_element fdef l1 x) l2 in
  let replace l1 l2 = rec_step (add_replacement fdef l1 repl) l2 in
  let cl_nil = nil_clause accum in
  let cls_many =
    if Relation.has_associative fdef.fd_relations
    then [many_clause fdef invg accum replace add_elem sd]
    else [] in
  let cls_single = single_clauses invg accum replace add_elem sd in
  let clauses = cl_nil :: (cls_many @ cls_single) in
  Match (arg, clauses)

and nil_clause accum = Clause (nil, list_rev accum)

and single_clauses invg accum replace add_elem =
  let left1 = replace_single_left invg accum replace in
  let left2 = replace_left_in_accum invg accum replace add_elem in
  let right1 = replace_single_right invg accum replace in
  let right2 = replace_right_in_accum invg accum replace add_elem in
  let default = default_clause accum add_elem in
  function
    | Left -> [left1; left2]
    | Right -> [right1; right2; default]
    | Both -> [left1; right2; right1; left2]

and replace_single_left invg =
  let inv x = pattern invg [x] in
  let pattern_left x y l = cons (inv x) (cons y l) in
  replace_single pattern_left

and replace_single_right invg =
  let inv x = pattern invg [x] in
  let pattern_right x y l = cons x (cons (inv y) l) in
  replace_single pattern_right

and replace_single pattern accum replace =
  Clause_when (pattern x y l, eq x y, replace accum l)

and replace_left_in_accum invg = 
  let p_inv x = pattern invg [x] in
  let pattern1 x l = cons x l in
  let constr_patt1 x = x in
  let pattern2 x l = cons (p_inv x) l in
  replace_in_accum pattern1 constr_patt1 pattern2

and replace_right_in_accum invg =
  let p_inv x = pattern invg [x] in
  let pattern1 x l = cons x l in
  let pattern2 x l = cons (p_inv x) l in
  let constr_patt2 x = construction1 invg x in
  replace_in_accum pattern2 constr_patt2 pattern1

and replace_in_accum pattern1 constr_patt1 pattern2 accum replace add_elem =
  Clause (pattern1 x l,
	  Match (accum, [
		   Clause_when (pattern2 y m, eq x y, replace m l);
		   Clause (underscore, add_elem (constr_patt1 x) accum l)
		 ]))

and many_clause fdef invg accum replace add_elem sd =
  let fg = fdef.fd_generator in
  let patt_inv_fg x = pattern invg [pattern fg [x]] in
  let inv_fg x = construction1 invg (construction1 fg x) in
  let default_expr = add_elem (inv_fg x) accum l in
  let right_side x n l = match sd with
    | Left -> replace_many_left x n l accum replace default_expr
    | Right -> replace_many_right x n l accum replace default_expr
    | Both -> replace_many_left x n l accum replace 
	(replace_many_right x n l accum replace default_expr) in
  Clause (cons (patt_inv_fg x) l, 
	  Let (ident_n, [], list_length x,
	        right_side x n l))

and replace_many_left x n l accum replace else_expr =
  If (bool_and (lt n (list_length l)) (eq x (take n l)),
      replace accum (drop n l),
      else_expr)

and replace_many_right x n l accum replace else_expr =
  If (bool_and (lt n (list_length accum)) (eq (list_rev x) (take n accum)),
      replace (drop n accum) l,
      else_expr)

and default_clause accum add_elem  =
    Clause (cons x l, add_elem x accum l)
;; 


let remove_consecutive_inverse fdef ident invg g sd l =
  let aux_ident = prefix_ident "aux" ident in
  Letrec (aux_ident, [accu; l],
	  aux_remove_consec_inverse fdef aux_ident accu l invg g sd,
	  apply2 aux_ident nil l)
;;

let genr_inverse fdef sd invg g =
  let fname = inverse_in fdef.fd_generator in
  let inverse_ident = prefix_ident "do" fname in
  let inverse_function =
    if Relation.has_commutative fdef.fd_relations
    then find_and_replace fdef inverse_ident
      (pattern1 invg x, construction1 invg x) (x, x) g
    else remove_consecutive_inverse fdef inverse_ident
      invg g sd in
  let main_function l =
    Letrec (inverse_ident, [l], inverse_function l,
	    element_if_empty g (apply1 inverse_ident l)) in
  ([fname, [l], main_function l],
   [apply1 fname])
;;

(* Distributivity *)

let rec genr_distrib_args c d e dist =
  let ident_distrib_args = distrib_args_in d in
  let clause_distrib_apply =
    if Check.is_varyadic_gen d then
      clause_dist_apply_v c d e dist
    else
      clause_dist_apply c d e dist in
  ident_distrib_args,
  [accu; l],
  Match (l, [
    clause_distrib_apply;
    clause_distrib_recurse ident_distrib_args;
    clause_distrib_not_found
  ])


(* C[xs, D(y1,...,yn), zs] -> E(C[xs,y1,zs], ..., C[xs,yn,zs]) *)
and clause_dist_apply c d e dist =
  match Check.arity_gen d with
    | 0 -> Clause (cons (pattern0 d) underscore, List [construction0 e])
    | n ->
	let xs = genr_args "x" n in
	let lhs = cons (pattern d xs) l
	and rhs = let1 ident_dist y (construction1 c (list_append accu (cons y l)))
	  (List [construction e (map dist (apply1 ident_dist) xs)]) in
	  Clause (lhs, rhs)

(* C[xs, D ys, zs] -> E [C[xs,y1,zs], ..., C[xs,yn,zs]] *)
and clause_dist_apply_v c d e dist =
  let lhs = cons (pattern1 d x) l
  and rhs = let1 ident_dist y (construction1 c (list_append accu (cons y l)))
    (List [construction1 e (apply2 (imap dist) (Constant ident_dist) x)]) in
  Clause (lhs, rhs)

and clause_distrib_recurse fun_ident =
  Clause (cons x l, apply2 fun_ident (cons x accu) l)

and clause_distrib_not_found = Clause (underscore, raise_not_found)
;;

(* Normalizing functions by relation *)

let genr_function_by_ap fdef rel =
  let ap = rel.prel_desc in
  let fg = fdef.fd_generator in
  match ap with
  | Commutative lidopt -> genr_sort fg lidopt
  | Associative _sd -> genr_flatten fg
  | Neutral (sd, g) ->
        genr_rem_neutral fdef sd g
  | Idempotent sd ->
	genr_idempotent fdef sd
  | Nilpotent (sd, g) ->
	genr_nilpotent fdef sd g
  | Inverse (sd, dname, None) ->
    (* We suppose here that a neutral element indeed exists for fdef.
       It should have been verified during the checking phase. *)
      let neutral_generator =
	match neutral fdef.fd_relations with
	  | Some (_ns, ng) -> ng
	  | None -> assert false in
      genr_inverse fdef sd dname neutral_generator
  | Inverse (sd, dname, Some dn) ->
      genr_inverse fdef sd dname dn
  | Distributive (_sd, g, hopt, dist) ->
      let h = match hopt with Some h -> h | None -> g
      and fname = distrib_in fg g in
      let genr_distrib_norm =
        fname,
        [x],
        Try (apply2 (distrib_args_in g) nil x, [Clause (exc_not_found, x)])
      in
	genr_distrib_args fdef.fd_generator g h dist :: [genr_distrib_norm],
        [apply1 fname]

  | Structure_item _ | Precedence _ | Status _ -> ([], [])
 
  (* Rewrite rules are handled separately from this function *)
  | Rewrite _ -> ([], [])

(* The following things do not make sense when used in a varyadic way
   See moca/compiler/ocaml_src/parsing/parsetree.mli for semantic hints
*)
  | (Division_by_Absorbent _
  | Absorbing _ (*FIXME?*)
  | Involutive) -> assert false
;;

let genr_code (prelude, norm) arg =
  let preamble = List.flatten prelude
  and normalize = List.fold_left (fun x f -> f x) arg (List.flatten norm)
  in (preamble, normalize)
;;

let common_aux_functions =
  [genr_mem; genr_delete_first; genr_unique_args; genr_take; genr_drop]
;;

let precedence rel =
  match rel.prel_desc with
  | Rewrite _ ->               10
  | Associative _ ->           20
  | Commutative _ ->           30
  | Division_by_Absorbent _
  | Neutral _
  | Involutive
  | Absorbing _
  | Idempotent _
  | Structure_item _
  | Status _
  | Precedence _ ->            40
  | Inverse _ ->               50
  | Nilpotent _ ->             60
  | Distributive _ ->          70
;;

let genr_function fdef arg =
  match fdef.fd_relations.prels_desc with
  | Prels_none | Prels_commented _ ->
    [], generator fdef.fd_generator [arg]
  | Prels_begend rls ->
      let rewrite_rels = Relation.rewrite fdef.fd_relations in
      let rewrite_code = genr_rewrite fdef rewrite_rels in
      let return_code = genr_return fdef.fd_generator rewrite_rels in
      let rls = List.sort (compare_relation precedence) rls in
      let codebyrel = List.map (genr_function_by_ap fdef) rls in
      let precode = List.split (rewrite_code :: codebyrel @ [return_code]) in
      genr_code precode arg
;;
