(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Fr��d��ric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2007,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: eqnrel.ml,v 1.34.2.1 2008/02/13 15:54:40 bonichon Exp $ *)

open Parsetree;;
open Code;;
open Term;;
open Term_utils;;
open Equation;;
open Useful;;
open Longident;;
open Var;;
open Symb;;

exception Not_algebraic;;


type ordering =
  | RPO | LPO | MPO
;;

let get_order, set_order =
  let order = ref LPO in
  (fun () -> !order),
  (fun o -> order := o)
;;


(** Creates a list of patterns from a pattern with or's. *)

let get_tuple p =
  match p.ppat_desc with
  | Ppat_tuple ps -> ps
  | Ppat_any | Ppat_var _ | Ppat_alias _ | Ppat_constant _
  | Ppat_construct _ | Ppat_variant _ | Ppat_record _ | Ppat_array _
  | Ppat_or _ | Ppat_constraint _ | Ppat_type _ | Ppat_parens _
    -> assert false
;;

let rec remove_or p =
  let loc d = { p with ppat_desc = d } in
  match p.ppat_desc with
  | Ppat_any
  | Ppat_construct (_, None, _)
  | Ppat_parens _
  | Ppat_tuple []
  | Ppat_var _ -> [p]

  | Ppat_alias (q, s) ->
      List.map (fun r -> loc (Ppat_alias (r, s))) (remove_or q)
  | Ppat_construct (lid, Some q, b) ->
      List.map (fun r -> loc (Ppat_construct (lid, Some r, b))) (remove_or q)
  | Ppat_constraint (q, ct) ->
      List.map (fun r -> loc (Ppat_constraint (r, ct))) (remove_or q)

  | Ppat_or (q, r) -> remove_or q @ remove_or r

  | Ppat_tuple (q :: qs) ->
      let ql = remove_or q
      and qsl = List.map get_tuple (remove_or (loc (Ppat_tuple qs))) in
      let f qi = List.map (fun qsj -> loc (Ppat_tuple (qi :: qsj))) qsl in
      List.flatten (List.map f ql)

  | Ppat_constant _
  | Ppat_variant _
  | Ppat_record _
  | Ppat_array _
  | Ppat_type _ -> raise Not_algebraic
;;

(** Convert an OCaml pattern into a term. *)

(* p must contain no or patterns *)
let term_of_pattern =

  let init, get_vmap, fresh, alias =
    let v = ref 0 and vmap = ref StrMap.empty in
      (fun () -> v := 0; vmap := StrMap.empty),
      (fun () -> !vmap),
      (fun s -> incr v; vmap := StrMap.add s (Var !v) !vmap; !v),
      (fun s t -> vmap := StrMap.add s t !vmap; t) in

  let rec term_of_pat p =
    match p.ppat_desc with
      | Ppat_any -> Var (fresh "_")
      | Ppat_var s -> Var (fresh s)
      | Ppat_alias (p, s) -> alias s (term_of_pat p)
      | Ppat_construct (lid, None, _b) -> iapp0 lid
      | Ppat_construct (lid, Some p, _b) -> iapp lid (terms_of_pat p)
      | Ppat_constraint (p, _ct) -> term_of_pat p
      | Ppat_parens p -> term_of_pat p
      | Ppat_or _ -> assert false
      | Ppat_tuple _
      | Ppat_type _
      | Ppat_constant _
      | Ppat_variant _
      | Ppat_record _
      | Ppat_array _ -> raise Not_algebraic

  and terms_of_pat p =
    match p.ppat_desc with
      | Ppat_any
      | Ppat_var _
      | Ppat_alias _
      | Ppat_construct _
      | Ppat_constraint _ -> [term_of_pat p]
      | Ppat_parens p -> terms_of_pat p
      | Ppat_tuple ps -> List.map term_of_pat ps
      | Ppat_or _ -> assert false
      | Ppat_type _
      | Ppat_constant _
      | Ppat_variant _
      | Ppat_record _
      | Ppat_array _ -> raise Not_algebraic in

    fun p ->
      let t = term_of_pat p in
      let vmap = get_vmap () in
	init ();
	t, vmap;;

(* could be optimized by computing the vmap only once ? *)
let terms_of_pattern p = List.map term_of_pattern (remove_or p);;

(** Convert an OCaml expression into a term. *)

let rec term_of_exp vmap e =
  match e.pexp_desc with
  | Pexp_ident ((Lident s) as lid) ->
      (try StrMap.find s vmap with Not_found -> iapp0 lid)
  | Pexp_ident ((Ldot _ | Lapply _) as lid) -> iapp0 lid
  | Pexp_construct (lid, None, _b) -> iapp0 lid
  | Pexp_construct (lid, Some e, _b) -> iapp lid (terms_of_exp vmap e)
  | Pexp_parens e
  | Pexp_begend e -> term_of_exp vmap e
  (* non algebraic expressions *)
  | Pexp_tuple _
  | Pexp_constant _ | Pexp_let _ | Pexp_function _ | Pexp_apply _
  | Pexp_match _ | Pexp_try _ | Pexp_variant _ | Pexp_record _
  | Pexp_field _ | Pexp_setfield _ | Pexp_array _ | Pexp_ifthenelse _
  | Pexp_sequence _ | Pexp_while _ | Pexp_for _ | Pexp_constraint _
  | Pexp_when _ | Pexp_send _ | Pexp_new _ | Pexp_setinstvar _
  | Pexp_override _ | Pexp_letmodule _ | Pexp_assert _
  | Pexp_assertfalse | Pexp_lazy _ | Pexp_poly _ | Pexp_object _
    -> raise Not_algebraic

and terms_of_exp vmap e =
  match e.pexp_desc with
  | Pexp_ident _
  | Pexp_construct _
  | Pexp_begend _ -> [term_of_exp vmap e]
  | Pexp_parens e -> terms_of_exp vmap e
  | Pexp_tuple es -> List.map (term_of_exp vmap) es
  (* non algebraic expressions *)
  | Pexp_constant _ | Pexp_let _ | Pexp_function _ | Pexp_apply _
  | Pexp_match _ | Pexp_try _ | Pexp_variant _ | Pexp_record _
  | Pexp_field _ | Pexp_setfield _ | Pexp_array _ | Pexp_ifthenelse _
  | Pexp_sequence _ | Pexp_while _ | Pexp_for _ | Pexp_constraint _
  | Pexp_when _ | Pexp_send _ | Pexp_new _ | Pexp_setinstvar _
  | Pexp_override _ | Pexp_letmodule _ | Pexp_assert _
  | Pexp_assertfalse | Pexp_lazy _ | Pexp_poly _ | Pexp_object _
    -> raise Not_algebraic
;;

(** Convert a Moca relation into a set of equations. *)

let eqns_of_rel c rel =
  match rel.prel_desc with
  | Rewrite (p, e) ->
      (try
	let l, vmap = term_of_pattern p in
	[l, term_of_exp vmap e]
      with Not_algebraic -> [])

  | Commutative _cmp -> [app2 c x y, app2 c y x]
  | Associative _sd -> [app2 c (app2 c x y) z, app2 c x (app2 c y z)]

  | Absorbing (Left, d) -> [app2 c (app2 d x y) y, y]
  | Absorbing (Right, d) -> [app2 c x (app2 d x y), x]
  | Absorbing (Both, d) ->
      [app2 c (app2 d x y) y, y; app2 c x (app2 d x y), x]

  | Neutral (Left, e) -> [app2 c (app0 e) x, x]
  | Neutral (Right, e) -> [app2 c x (app0 e), x]
  | Neutral (Both, e) ->
      [app2 c (app0 e) x, x; app2 c x (app0 e), x]

  | Idempotent _ when Check.arity_gen c = 1 -> [app1 c (app1 c x), app1 c x]
  | Idempotent Left -> [app2 c x (app2 c x y), app2 c x y]
  | Idempotent Right -> [app2 c (app2 c x y) y, app2 c x y]
  | Idempotent Both ->
      [app2 c x (app2 c x y), app2 c x y; app2 c (app2 c x y) y, app2 c x y]

  | Nilpotent (_, a) when Check.arity_gen c = 1 -> [app1 c (app1 c x), app0 a]
  | Nilpotent (Left, a) -> [app2 c x (app2 c x y), app2 c (app0 a) y]
  | Nilpotent (Right, a) -> [app2 c (app2 c x y) y, app2 c x (app0 a)]
  | Nilpotent (Both, a) ->
      [app2 c x (app2 c x y), app2 c (app0 a) y;
       app2 c (app2 c x y) y, app2 c x (app0 a)]

  | Involutive -> [app1 c (app1 c x), x]

  | Inverse (Left, i, None) ->
      let e = Check.neutral_element_gen c in [app2 c (app1 i x) x, app0 e]
  | Inverse (Right, i, None) ->
      let e = Check.neutral_element_gen c in [app2 c x (app1 i x), app0 e]
  | Inverse (Both, i, None) ->
      let e = Check.neutral_element_gen c in
      [app2 c (app1 i x) x, app0 e; app2 c x (app1 i x), app0 e]

  | Inverse (Left, i, Some a) -> [app2 c (app1 i x) x, app0 a]
  | Inverse (Right, i, Some a) -> [app2 c x (app1 i x), app0 a]
  | Inverse (Both, i, Some a) ->
      [app2 c (app1 i x) x, app0 a; app2 c x (app1 i x), app0 a]

  | Distributive (_, d, gopt, b) when Check.arity_gen c = 1 ->
      let e = match gopt with Some e -> e | None -> d
      and n = Check.arity_gen d in
      let xs = vars n in
      let args = List.map (app1 c) xs in
      let args = match b with
	| Dist_Inverse -> List.rev args
	| Dist_Direct -> args in
      [app1 c (app d xs), app e args]
  | Distributive (Left, d, gopt, b) ->
      let e = match gopt with Some e -> e | None -> d
      and n = Check.arity_gen d in
      let xs = vars n in
      let args = List.map (fun x -> app2 c x z) xs in
      let args = match b with
	| Dist_Inverse -> List.rev args
	| Dist_Direct -> args in
      [app2 c (app d xs) z, app e args]
  | Distributive (Right, d, gopt, b) ->
      let e = match gopt with Some e -> e | None -> d
      and n = Check.arity_gen d in
      let xs = vars n in
      let args = List.map (fun x -> app2 c z x) xs in
      let args = match b with
	| Dist_Inverse -> List.rev args
	| Dist_Direct -> args in
      [app2 c z (app d xs), app e args]
  | Distributive (Both, d, gopt, b) ->
      let e = match gopt with Some e -> e | None -> d
      and n = Check.arity_gen d in
      let xs = vars n in
      let args1 = List.map (fun x -> app2 c x z) xs
      and args2 = List.map (fun x -> app2 c z x) xs in
      let args1, args2 = match b with
	| Dist_Inverse -> List.rev args1, List.rev args2
	| Dist_Direct -> args1, args2 in
      [app2 c (app d xs) z, app e args1;
       app2 c z (app d xs), app e args2]

  | Division_by_Absorbent _
  | Structure_item _
  | Status _
  | Precedence _ -> []
;;

let eqnset_of_rels c rels =
  match rels.prels_desc with
    | Prels_none
    | Prels_commented _ -> EqnSet.empty
    | Prels_begend rs ->
	EqnSet.union_map (fun r -> EqnSet.of_list (eqns_of_rel c r)) rs
;;

let eqnset_of_cd (cname, _cts, rels, _loc) =
  let c = { pgen_desc = Lident cname; pgen_loc = Location.none } in
    eqnset_of_rels c rels
;;

let eqnset_of_cds = EqnSet.union_map eqnset_of_cd
;;

let eqnset_of_ntd (_tname, td) =
  match td.ptype_kind with
    | Ptype_variant cds -> eqnset_of_cds cds
    | Ptype_record _
    | Ptype_abstract _ -> EqnSet.empty
;;

let eqnset_of_ntds = EqnSet.union_map eqnset_of_ntd
;;

(** Convert a term into a Code.exp. *)

let exp_of_var x = make_Constant (string_of_var x);;

let rec exp_of_term = function
  | Var x -> exp_of_var x
  | App (f, []) -> make_Constant (string_of_symbol f)
  | App (f, ts) ->
      Apply (make_Constant (string_of_symbol f), List.map exp_of_term ts)
;;

let pattern_of_term = exp_of_term;;

(** Convert a rule into a Code.clause. *)

let linearize =

  let init, get_vmap, fresh =
    let v = ref 0 and vset = ref IntSet.empty and vmap = ref IntMap.empty in
      (fun () -> v := 0; vset := IntSet.empty; vmap := IntMap.empty),
      (fun () -> !vmap),
      (fun x ->
        if IntSet.mem x !vset
        then (incr v; vmap := IntMap.add !v x !vmap; Var !v)
        else (vset := IntSet.add x !vset; Var x)) in

  let rec aux = function
    | Var x -> fresh x
    | App (f, ts) -> App (f, List.map aux ts) in

    fun t ->
      let t = aux t in
      let vmap = get_vmap () in
	init ();
	t, vmap
;;

let rename vmap =
  let rec aux = function
    | Var x as t -> (try Var (IntMap.find x vmap) with Not_found -> t)
    | App (f, ts) -> App (f, List.map aux ts)
  in aux
;;
 
let clause_of_rule (l, r) =
  let l, vmap = linearize l in
  let bool_exp_of_mapping v x e =
    infix_name e "&&"
      (Genr_base.eq_function (exp_of_var v) (exp_of_var x))
  and id_true = make_Constant "true" in
  let condition = IntMap.fold bool_exp_of_mapping vmap id_true in
    if condition == id_true
    then Clause (exp_of_term l, exp_of_term r)
    else Clause_when (exp_of_term l, condition, exp_of_term r)
;;

(** Build a precedence from the precedence declarations.
    Add RPO status information.
 *)

open Prec;;

let h_rpostatus = Hashtbl.create 7;;

let insert ((_c1,  k1) as x1) =
  let rec aux = function
    | [] -> [x1]
    | (_c2,  k2) as x2 :: l when k1 > k2 -> x2 :: aux l
    | l -> x1 :: l
  in aux
;;
 
(* Default for status when nothing is given:
   if symbol is commutative, status = mul
   otherwise lex
*)

let rec add_rel cmap cname r =
  match r.prel_desc with
    | Precedence k -> (insert (cname, k) cmap)
      (* No need to order status *)
    | Status s -> (Hashtbl.add h_rpostatus
                  (symbol_of_string cname) s;
                  cmap)                                                    
    | Commutative _ -> 
      (let sname = symbol_of_string cname in
      if not (Hashtbl.mem h_rpostatus sname) then
       Hashtbl.add h_rpostatus sname Mul;
      cmap)
    | Rewrite _ | Associative _
    | Absorbing _ | Neutral _ | Idempotent _
    | Nilpotent _ | Involutive | Inverse _ | Distributive _
    | Division_by_Absorbent _ | Structure_item _ -> cmap
;;

let rec add_rel_list cmap cname = function
  | [] -> cmap
  | r :: rs -> add_rel_list (add_rel cmap cname r) cname rs
;;

let add_rels cmap cname rels =
  match rels.prels_desc with
    | Prels_none
    | Prels_commented _ -> cmap
    | Prels_begend rs -> add_rel_list cmap cname rs
;;

let rec add_cds cmap = function
  | [] -> cmap
  | (cname, _cts, rels, _loc) :: cds -> add_cds (add_rels cmap cname rels) cds
;;

let add_ntd cmap (_tname, td) =
  match td.ptype_kind with
    | Ptype_record _
    | Ptype_abstract _ -> cmap
    | Ptype_variant cds -> add_cds cmap cds
;;

let rec add_ntds cmap = function
  | [] -> cmap
  | ntd :: ntds -> add_ntds (add_ntd cmap ntd) ntds
;;


let prec_status_of_ntds ntds =
  let precs = add_ntds [] ntds in
  prec_of_list (List.map (fun (x, _) -> symbol_of_string x) precs)
                  
;;

(** Completion. *)

open Order;;
open Comp;;
open Rule;;
open Configuration;;  

let completion n ntds =
  let prec = prec_status_of_ntds ntds
  and es = eqnset_of_ntds ntds
  and acs = RulSet.empty in
  let ordering =
    match get_order () with
      | LPO -> lpo
      | MPO -> mpo
      | RPO -> rpo h_rpostatus
  in
  let gt t u = ordering (Prec.compare prec) t u = Greater in
    Printf.printf "Equations:\n%a\n" (to_channel EqnSet.bprint) es;
    Debug.verbose := true;
    Debug.print_rules := false;
    Debug.print_stats := true;
    Debug.trace "step";
    Debug.trace "rule";
    Debug.untrace "cp";
    try
      let ((es,rs) as ers), i = complete_n acs gt n (es,RulSet.empty) in
	Debug.sep ();
	if EqnSet.is_empty es then
	  Printf.printf "Completed in %i steps:\n%a\n"
	    i (to_channel RulSet.bprint) rs
	else if i = n then
	  Printf.printf "Incompleted after %i steps:\n%a\n"
	    i (to_channel Comp.bprint) ers
    with Unorientable es ->
      Debug.sep (); Printf.printf "Unorientable equations:\n%a\n"
	(to_channel EqnSet.bprint) es
;;
