(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Laura Lowenthal, projet Protheo, INRIA Lorraine           *)
(*                                                                    *)
(*  Copyright 2005-2007,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: genr_equalities.ml,v 1.15 2008/02/12 17:48:34 lowenthl Exp $ *)

(** {Generation of equalities on values, deduced from moca relations.} *)

open Funutils
open Eqnrel
open Code
open Genr_base
open Tgenr_base
open Typed_vars
open Genr_substitutions
open Symb
;;

let print_no_equalities_warning number exp =
  let ob = Buffer.create 128 in
    Printf.bprintf ob "Warning: cannot test equation %a: \
      could not generate enough values for %n equations\n"
      Pr_code.exp exp number;
    prerr_string (Buffer.contents ob);
    Buffer.clear ob 

let print_unknown_type_warning exp =
  let ob = Buffer.create 128 in
    Printf.bprintf ob "Warning: cannot test equation %a: \
      could not assign types to variables.\n"
      Pr_code.exp exp;
    prerr_string (Buffer.contents ob);
    Buffer.clear ob
      
let rec exp_subst_of_term subst term = match term with
  | Term.Var x -> subst x
  | Term.App (x, ts) ->
      let string = string_of_symbol x in
      match ts with
       | [t1; t2] when Pr_otype.is_infix string ->
          infix_name (exp_subst_of_term subst t1) string
           (exp_subst_of_term subst t2)
       | _ -> mk_construction (string_of_symbol x) ^$
               List.map (exp_subst_of_term subst) ts
;;

let exp_of_term = exp_subst_of_term (make_Constant <.> Var.string_of_var)
;;

let exp_subst_of_equation subst (left_term, right_term) =
  mk_equality (exp_subst_of_term subst left_term)
   (exp_subst_of_term subst right_term)
;;

let exp_of_equation (left_term, right_term) =
  mk_equality (exp_of_term left_term) (exp_of_term right_term)
;;

let rec add_val_defs final_exp vars subst =
  match vars with
    [] -> final_exp
  | (v, _) :: vs ->
     let v_ident = Longident.Lident (Var.string_of_var v) in
     Code.let0 v_ident (subst v) (add_val_defs final_exp vs subst)

let equations gen rels =
    if Check.is_varyadic_gen gen
    then Genr_varyadic_equations.testing_equations_of_rels gen rels
    else Equation.EqnSet.elements (eqnset_of_rels gen rels)
;;

let genr_values_and_equalities_for_eqn number f_gen_values eqn =
  let exp = exp_of_equation eqn in
  try
    let vars = typed_vars_eqn eqn in
    let substs = genr_substs number f_gen_values vars in
    List.map (add_val_defs exp vars) substs
  with
    | Not_enough_values ->
	print_no_equalities_warning number exp; []
    | Unknown_type ->
	print_unknown_type_warning exp; []
;;

let genr_values_and_equalities number f_gen_values (name, _, rels, loc) =
  let generator = mk_generator name loc in
  let eqns = equations generator rels in
  let gen_values_and_eqs =
    genr_values_and_equalities_for_eqn number f_gen_values in
  Listutils.flat_map gen_values_and_eqs eqns
;;
